import abc
import asyncio
import contextlib
import sys
from functools import partial
from itertools import chain
from typing import Sequence, Optional, Dict

import networkx as nx

import statey as st
from statey import exc
from statey.plan.actions import (
    ExecuteTaskSession,
    SetValue,
    DeleteValue,
    ResourceSetValue,
)
from statey.plan.plans import Plan, DefaultPlan, PlanNode, PlanNodeState, StatefulPlan
from statey.provider import Provider, ProviderId
from statey.resource import (
    ResourceSession,
    ResourceGraph,
    StateConfig,
    StateSnapshot,
    DefaultResourceGraph,
)
from statey.syms import Object, stack, types
from statey.task import TaskSession, create_task_session


class Migrator(abc.ABC):
    """
    Migrator is the interface for creating plans for migrating one set of resource states
    to another
    """

    @abc.abstractmethod
    async def plan(
        self,
        config_session: ResourceSession,
        state_graph: Optional[ResourceGraph] = None,
        state_session: Optional[ResourceSession] = None,
    ) -> Plan:
        """
        Create a task graph to create the resource states in config_session, migrating
        from state_session if it is passed
        """
        raise NotImplementedError


class DefaultMigrator(Migrator):
    """
    simple default Migrator implementation
    """

    @contextlib.contextmanager
    def error_ctx(self) -> None:
        with stack.rewrite_ctx():
            try:
                yield
            except Exception as err:
                raise exc.ErrorDuringPlanning(err, sys.exc_info()[2]) from err

    @contextlib.contextmanager
    def node_error_ctx(
        self, key: str, snapshot: StateSnapshot, config: StateConfig
    ) -> None:
        with stack.rewrite_ctx():
            try:
                yield
            except Exception as err:
                raise exc.ErrorDuringPlanningNode(
                    key, snapshot, config, err, sys.exc_info()[2]
                ) from err

    def create_task_session(
        self, data: Optional[Dict[str, Object]] = None
    ) -> TaskSession:
        if data is None:
            data = {}
        session = create_task_session()
        for key, val in data.items():
            session.set(key, val)
        return session

    async def get_or_create_provider(
        self,
        id: ProviderId,
        providers: Dict[ProviderId, Provider],
        registry: "Registry",
    ) -> Provider:
        """"""
        provider = providers.get(id)
        if provider is None:
            provider = providers[id] = registry.get_provider(*id)
            await provider.setup()

        return provider

    async def plan_node_migrate_states(
        self,
        node: str,
        config_session: ResourceSession,
        state_session: ResourceSession,
        config_dep_graph: nx.DiGraph,
        state_dep_graph: nx.DiGraph,
        output_session: ResourceSession,
        providers: Dict[ProviderId, Provider],
    ) -> Sequence[PlanNode]:
        """"""
        config_bound_state = output_session.get_state(node)
        config_state = config_bound_state.state
        previous_state = state_session.get_state(node).state

        current_depends_on = list(state_dep_graph.pred[node])
        config_depends_on = list(config_dep_graph.pred[node])

        provider = await self.get_or_create_provider(
            config_state.provider, providers, config_session.ns.registry
        )

        resource = provider.get_resource(config_state.resource)

        previous_ref = state_session.ns.ref(node)
        previous_resolved = state_session.resolve(
            previous_ref, decode=False, allow_unknowns=True
        )
        config_partial_resolved = output_session.resolve(
            config_bound_state.input, decode=False, allow_unknowns=True
        )

        task_session = self.create_task_session(
            {
                "input": Object[config_state.input_type](config_partial_resolved),
                "state": Object[previous_state.output_type](previous_resolved),
            }
        )

        input_ref = task_session.ns.ref("input")
        task_state_ref = task_session.ns.ref("state")

        previous_snapshot = StateSnapshot(task_state_ref, previous_state)
        bound_config = StateConfig(input_ref, config_state)

        current_action = None
        state_symbol = previous_ref

        with self.node_error_ctx(node, previous_snapshot, bound_config):

            try:
                plan_output = await resource.plan(
                    previous_snapshot, bound_config, task_session
                )
            except exc.NullRequired:
                null_bound_config = StateConfig({}, resource.null_state)

                old_task_session = self.create_task_session(
                    {
                        "input": Object[resource.null_state.input_type]({}),
                        "state": Object[previous_state.output_type](previous_resolved),
                    }
                )
                old_state_ref = old_task_session.ns.ref("state")

                old_previous_snapshot = StateSnapshot(old_state_ref, previous_state)

                # old_task_session = task_session_base.clone()
                og_plan_output = await resource.plan(
                    old_previous_snapshot, null_bound_config, old_task_session
                )

                if isinstance(og_plan_output, tuple) and len(og_plan_output) == 2:
                    og_output_ref, og_graph_ref = og_plan_output
                else:
                    og_output_ref, og_graph_ref = og_plan_output, og_plan_output

                null_snapshot = StateSnapshot({}, resource.null_state)

                task_session = self.create_task_session(
                    {
                        "input": Object[config_state.input_type](
                            config_partial_resolved
                        ),
                        "state": Object[resource.null_state.output_type]({}),
                    }
                )

                plan_output = await resource.plan(
                    null_snapshot, bound_config, task_session
                )

                current_action = ExecuteTaskSession(
                    task_session=old_task_session,
                    task_input_key="input",
                    task_state_key="state",
                    output_key=node,
                    output_ref=og_output_ref,
                    graph_ref=og_graph_ref,
                    config_state=resource.null_state,
                    previous_state=previous_state,
                    resource=resource,
                    input_symbol=null_snapshot.obj,
                    state_symbol=previous_ref,
                )
                state_symbol = Object({})

        # Allow different refs for graph vs. session by returning a tuple of refs
        if isinstance(plan_output, tuple) and len(plan_output) == 2:
            output_ref, graph_ref = plan_output
        else:
            output_ref, graph_ref = plan_output, plan_output

        output_partial_resolved = task_session.resolve(
            output_ref, allow_unknowns=True, decode=False
        )
        output_session.set_data(node, output_partial_resolved)

        test_graph = task_session.task_graph(DefaultResourceGraph(), node)
        # This indicates there are no tasks in the session.
        if not test_graph.nodes:
            try:
                resolved_graph_ref = task_session.resolve(graph_ref, decode=False)
            # The graph reference isn't fully resolved yet, we still need
            # to execute the task session
            except exc.UnknownError:
                pass
            else:
                state_obj = Object(
                    resolved_graph_ref,
                    config_state.state.output_type,
                    config_session.ns.registry,
                )
                # Since we are not setting this key in the graph, we can
                # act like the configuration has no dependencies during
                # planning. This essentially lets us avoid circular
                # dependencies because these no-op resources will never
                # depend on anything.
                plan_node = PlanNode(
                    key=node,
                    current=PlanNodeState(
                        data=previous_resolved,
                        type=previous_state.state.output_type,
                        state=previous_state,
                        depends_on=current_depends_on,
                        action=None,
                    ),
                    config=PlanNodeState(
                        data=resolved_graph_ref,
                        type=config_state.state.output_type,
                        state=config_state,
                        depends_on=config_depends_on,
                        action=ResourceSetValue(
                            node, state_obj, resource, config_state
                        ),
                    ),
                    enforce_dependencies=False,
                )
                return [plan_node]

        action = ExecuteTaskSession(
            task_session=task_session,
            task_input_key="input",
            task_state_key="state",
            output_key=node,
            output_ref=output_ref,
            graph_ref=graph_ref,
            config_state=config_state,
            previous_state=previous_state,
            resource=resource,
            input_symbol=config_bound_state.input,
            state_symbol=state_symbol,
        )

        plan_node = PlanNode(
            key=node,
            current=PlanNodeState(
                data=previous_resolved,
                type=previous_state.state.output_type,
                state=previous_state,
                depends_on=current_depends_on,
                action=current_action,
            ),
            config=PlanNodeState(
                data=output_partial_resolved,
                type=config_state.state.output_type,
                state=config_state,
                depends_on=config_depends_on,
                action=action,
            ),
        )
        return [plan_node]

    # async def plan_node_previous(
    #     self,
    #     node: str,
    #     config_session: ResourceSession,
    #     state_session: ResourceSession,
    #     config_dep_graph: nx.DiGraph,
    #     state_dep_graph: nx.DiGraph,
    #     output_session: ResourceSession,
    #     providers: Dict[ProviderId, Provider],
    # ) -> Sequence[PlanNode]:

    #     pass

    # async def plan_node_config(
    #     self,
    #     node: str,
    #     config_session: ResourceSession,
    #     state_session: ResourceSession,
    #     config_dep_graph: nx.DiGraph,
    #     state_dep_graph: nx.DiGraph,
    #     output_session: ResourceSession,
    #     providers: Dict[ProviderId, Provider],
    # ) -> Sequence[PlanNode]:

    #     pass

    async def plan_node(
        self,
        node: str,
        config_session: ResourceSession,
        state_session: ResourceSession,
        config_dep_graph: nx.DiGraph,
        state_dep_graph: nx.DiGraph,
        output_session: ResourceSession,
        providers: Dict[ProviderId, Provider],
    ) -> Sequence[PlanNode]:
        """"""
        previous_exists = node in state_dep_graph.nodes
        previous_state = None
        previous_bound_state = None
        try:
            previous_bound_state = state_session.get_state(node)
        except exc.SymbolKeyError:
            pass
        else:
            previous_state = previous_bound_state.state

        config_exists = node in config_dep_graph.nodes
        config_bound_state = None
        config_state = None
        try:
            config_bound_state = output_session.get_state(node)
        except exc.SymbolKeyError:
            pass
        else:
            config_state = config_bound_state.state

        current_depends_on = list(state_dep_graph.pred[node]) if previous_exists else []
        config_depends_on = list(config_dep_graph.pred[node]) if config_exists else []

        plan_nodes = []

        # We can handle this node with one action--otherwise we need two, though
        # one or both may just be state update operations.
        if (
            config_state is not None
            and previous_state is not None
            and config_state.resource == previous_state.resource
        ):
            return await self.plan_node_migrate_states(
                node=node,
                config_session=config_session,
                state_session=state_session,
                config_dep_graph=config_dep_graph,
                state_dep_graph=state_dep_graph,
                output_session=output_session,
                providers=providers,
            )

        current_node_state = None
        config_node_state = None

        if previous_state:
            provider = await self.get_or_create_provider(
                previous_state.provider, providers, state_session.ns.registry
            )

            resource = provider.get_resource(previous_state.resource)

            task_session = self.create_task_session()

            previous_ref = state_session.ns.ref(node)
            previous_resolved = state_session.resolve(
                previous_ref, decode=False, allow_unknowns=True
            )

            task_session = self.create_task_session(
                {
                    "state": Object[previous_state.output_type](previous_resolved),
                    "input": Object[resource.null_state.input_type]({}),
                }
            )

            state_ref = task_session.ns.ref("state")
            previous_snapshot = StateSnapshot(state_ref, previous_state)

            input_ref = task_session.ns.ref("input")
            config_bound = StateConfig(input_ref, resource.null_state)

            with self.error_ctx():
                plan_output = await resource.plan(
                    previous_snapshot, config_bound, task_session
                )

            # Allow different refs for graph vs. session by returning a tuple of refs
            if isinstance(plan_output, tuple) and len(plan_output) == 2:
                output_ref, graph_ref = plan_output
            else:
                output_ref, graph_ref = plan_output, plan_output

            with self.error_ctx():
                # In theory this should _not_ allow unknowns, but keeping it less strict for now.
                # Update: removed unknowns, they should not be needed
                # Update2: These are needed for now
                output_resolved = task_session.resolve(
                    output_ref, decode=False, allow_unknowns=True
                )

            action = ExecuteTaskSession(
                task_session=task_session,
                task_input_key="input",
                task_state_key="state",
                output_key=node,
                output_ref=output_ref,
                graph_ref=graph_ref,
                config_state=resource.null_state,
                previous_state=previous_state,
                resource=resource,
                input_symbol=Object({}, types.EmptyType()),
                state_symbol=previous_ref,
            )

            current_node_state = PlanNodeState(
                data=previous_resolved,
                type=previous_state.output_type,
                state=previous_state,
                action=action,
                depends_on=current_depends_on,
            )
        elif previous_exists and not config_exists:
            current_node_state = PlanNodeState(
                data=state_dep_graph.nodes[node]["value"],
                type=state_dep_graph.nodes[node]["type"],
                state=None,
                action=DeleteValue(node),
                # If the old value is not stateful, there are no dependencies to deleting it.
                depends_on=[],
            )

        if config_state:
            provider = providers.get(config_state.provider)
            if provider is None:
                provider = providers[
                    config_state.provider
                ] = config_session.ns.registry.get_provider(*config_state.provider)
                await provider.setup()

            resource = provider.get_resource(config_state.resource)
            config_partial_resolved = output_session.resolve(
                config_bound_state.input, decode=False, allow_unknowns=True
            )

            task_session = self.create_task_session(
                {
                    "state": Object[resource.null_state.output_type]({}),
                    "input": Object[config_state.state.input_type](
                        config_partial_resolved
                    ),
                }
            )

            state_ref = task_session.ns.ref("state")
            previous_snapshot = StateSnapshot(state_ref, resource.null_state)

            input_ref = task_session.ns.ref("input")
            config_bound = StateConfig(input_ref, config_state)

            with self.error_ctx():
                plan_output = await resource.plan(
                    previous_snapshot, config_bound, task_session
                )

            # End error junk
            # Allow different refs for graph vs. session by returning a tuple of refs
            if isinstance(plan_output, tuple) and len(plan_output) == 2:
                output_ref, graph_ref = plan_output
            else:
                output_ref, graph_ref = plan_output, plan_output

            output_partial_resolved = task_session.resolve(
                output_ref, allow_unknowns=True, decode=False
            )

            output_session.set_data(node, output_partial_resolved)

            action = ExecuteTaskSession(
                task_session=task_session,
                task_input_key="input",
                task_state_key="state",
                output_key=node,
                output_ref=output_ref,
                graph_ref=graph_ref,
                config_state=config_state,
                previous_state=resource.null_state,
                resource=resource,
                input_symbol=config_bound_state.input,
                state_symbol=Object({}),
            )

            config_node_state = PlanNodeState(
                data=output_partial_resolved,
                type=config_state.state.output_type,
                state=config_state,
                action=action,
                depends_on=config_depends_on,
            )
        elif config_exists:
            ref = config_session.ns.ref(node)

            config_node_state = PlanNodeState(
                data=config_session.resolve(ref, decode=False, allow_unknowns=True),
                type=config_session.ns.resolve(node),
                state=None,
                action=SetValue(node, ref),
                depends_on=config_depends_on,
            )

        plan_node = PlanNode(node, current_node_state, config_node_state)
        plan_nodes.append(plan_node)

        return plan_nodes

    async def plan(
        self,
        config_session: ResourceSession,
        state_graph: Optional[ResourceGraph] = None,
        state_session: Optional[ResourceSession] = None,
    ) -> Plan:

        from statey.executor import AsyncIOTaskGraph, AsyncIOGraphExecutor

        if state_graph is None:
            state_graph = DefaultResourceGraph()

        if state_session is None:
            state_session = state_graph.to_session()

        with self.error_ctx():
            config_dep_graph = config_session.dependency_graph()

        with self.error_ctx():
            state_dep_graph = state_session.dependency_graph()

        # For planning we need a copy of the config session that we will partially resolve
        # (maybe w/ unknowns) as we go
        output_session = config_session.clone()

        config_nodes = set(config_dep_graph.nodes)
        state_nodes = set(state_dep_graph.nodes)

        state_only_nodes = state_nodes - config_nodes

        plan_nodes = []
        providers = {}

        task_graph = nx.DiGraph()

        async def coro(node):
            plan_nodes.extend(
                await self.plan_node(
                    node=node,
                    config_session=config_session,
                    state_session=state_session,
                    config_dep_graph=config_dep_graph,
                    state_dep_graph=state_dep_graph,
                    output_session=output_session,
                    providers=providers,
                )
            )

        for node in chain(config_dep_graph.nodes, state_only_nodes):
            task_graph.add_node(node, coroutine_factory=partial(coro, node))

        task_graph.add_edges_from(((a, b) for a, b, _ in config_dep_graph.edges))
        task_graph_obj = AsyncIOTaskGraph(task_graph)
        executor = AsyncIOGraphExecutor()

        try:
            exec_info = await executor.execute_async(task_graph_obj)
        finally:
            await asyncio.gather(
                *(provider.teardown() for provider in providers.values())
            )

        with self.error_ctx():
            exec_info.raise_for_failure()

        config_out = config_session.clone()

        # Before creating the plan, iterate through the output session and set
        # each value as an expectation for the config session
        for key in output_session.ns.keys():
            ref = output_session.ns.ref(key)
            value = output_session.resolve(ref, allow_unknowns=True, decode=False)
            current_value = output_session.get_data(key)
            typ = output_session.ns.resolve(key)
            obj = Object(current_value, typ) >> value
            config_out.set_data(key, obj)

        plan = DefaultPlan(
            nodes=tuple(plan_nodes),
            providers=providers,
            config_session=config_out,
            state_session=state_session,
            state_graph=state_graph,
            migrator=self,
        )
        # This will ensure that the task graph is a valid DAG, not perfect but it will
        # do for now
        plan.build_task_graph()
        return plan


class StatefulMigrator(DefaultMigrator):
    """
    Migrator that wraps the plan in a StatefulPlan and accepts an addition "state" keyword
    argument to the plan() method.
    """

    async def plan(
        self,
        config_session: ResourceSession,
        state_graph: Optional[ResourceGraph] = None,
        state_session: Optional[ResourceSession] = None,
        *,
        state: str
    ) -> Plan:

        plan = await super().plan(config_session, state_graph, state_session)
        return StatefulPlan(plan, state, self)



import abc
import dataclasses as dc
from functools import reduce
from itertools import product
from typing import Sequence, Optional, Dict, Tuple, Any, Callable

import networkx as nx

from statey.plan.actions import PlanAction
from statey.provider import Provider
from statey.resource import ResourceState, ResourceSession, ResourceGraph
from statey.syms import types, session
from statey.task import StateSnapshot, SnapshotSwitch, SetGraphState


def check_dag(graph: nx.DiGraph, strict: bool = False) -> None:
    """
    Check that this graph is a DAG, attempting to remove optional
    edges to remove conflicts if necessary
    """
    while not strict:
        try:
            cycle = nx.find_cycle(graph, orientation="reverse")
        except nx.NetworkXNoCycle:
            return

        removed = False

        for left, right, _ in cycle:
            edge = graph.edges[left, right]
            if edge.get("optional"):
                graph.remove_edge(left, right)
                removed = True
                break

        if not removed:
            break

    # This will raise a properly formatted error
    utils.check_dag(dag)


@dc.dataclass(frozen=True)
class PlanNodeState:
    """
    Encapsulates information about one the key associated with a plan node, and
    the transition required to get there
    """

    data: Any
    type: types.Type
    state: Optional[ResourceState]
    depends_on: Sequence[str]
    action: PlanAction


@dc.dataclass(frozen=True)
class PlanNode:
    """
    A plan node contains information about the state of a given name and before
    and after this migration, including any unknowns. It also contains information
    about the up and downstream dependencies of the configuration and previous state
    respectively
    """

    key: str
    current: Optional[PlanNodeState] = None
    config: Optional[PlanNodeState] = None
    enforce_dependencies: bool = True

    def input_task(self) -> Optional[str]:
        """"""
        return self.current_input_task() or self.config_input_task()

    def output_task(self) -> Optional[str]:
        """"""
        return self.config_output_task() or self.current_output_task()

    def current_prefix(self) -> str:
        """"""
        if (
            self.current is not None
            and self.current.action is not None
            and self.config is not None
            and self.config.action is not None
        ):
            return f"{self.key}:current"
        return self.key

    def config_prefix(self) -> str:
        """"""
        if (
            self.current is not None
            and self.current.action is not None
            and self.config is not None
            and self.config.action is not None
        ):
            return f"{self.key}:config"
        return self.key

    def current_output_task(self) -> Optional[str]:
        """
        The output task of migrating the current state. This may or may not be the
        same as output_task(); the case where it is the same is when either this key
        only exists in one of the previous of configured namespaces or when it is a single
        resource whose state is being migrated.
        """
        action = self.current and self.current.action
        if action is None:
            return None
        return action.output_task(self.current_prefix())

    def config_output_task(self) -> Optional[str]:
        """"""
        action = self.config and self.config.action
        if action is None:
            return None
        return action.output_task(self.config_prefix())

    def config_input_task(self) -> str:
        """"""
        action = self.config and self.config.action
        if action is None:
            return None
        return action.input_task(self.config_prefix())

    def current_input_task(self) -> str:
        """"""
        action = self.current and self.current.action
        if action is None:
            return None
        return action.input_task(self.current_prefix())

    def get_edges(
        self, other_nodes: Dict[str, "PlanNode"]
    ) -> Sequence[Tuple[str, str]]:
        """
        Get the edges for this node, given the other nodes.
        """
        edges = {}

        if not self.enforce_dependencies:
            return []

        if self.input_task() is None:
            return []

        config_input_task = self.config_input_task()
        current_output_task = self.current_output_task() or self.output_task()

        for upstream in self.config and self.config.depends_on or []:
            node = other_nodes[upstream]
            ref = node.output_task()
            if ref is None or config_input_task is None:
                continue
            edges[ref, config_input_task] = {"optional": False}

        for downstream in self.current and self.current.depends_on or []:
            node = other_nodes[downstream]
            input_ref = node.input_task()
            if input_ref is None or current_output_task is None:
                continue
            edges[current_output_task, input_ref] = {"optional": True}

        return [(*key, val) for key, val in edges.items()]

    def get_task_graph(
        self,
        resource_graph: ResourceGraph,
        output_session: session.Session,
        state_session: session.Session,
    ) -> nx.DiGraph:
        """
        Render a task graph for this specific node of the plan. The names specified
        by current_output_task() and output_task() should be tasks in this graph.
        """
        tasks = nx.DiGraph()

        if (
            self.current
            and self.current.action is not None
            and self.config
            and self.config.action is not None
        ):
            self.current.action.render(
                graph=tasks,
                prefix=self.current_prefix(),
                resource_graph=resource_graph,
                output_session=output_session,
                state_session=state_session,
                # The current task will always have a null result
                dependencies=(),
            )
            self.config.action.render(
                graph=tasks,
                prefix=self.config_prefix(),
                resource_graph=resource_graph,
                output_session=output_session,
                dependencies=self.config.depends_on,
                state_session=state_session,
            )
            # Make these two graphs depend on one another
            tasks.add_edge(self.current_output_task(), self.config_input_task())
        elif self.current and self.current.action:
            self.current.action.render(
                graph=tasks,
                prefix=self.current_prefix(),
                resource_graph=resource_graph,
                output_session=output_session,
                state_session=state_session,
                dependencies=(),
            )
        elif self.config and self.config.action:
            self.config.action.render(
                graph=tasks,
                prefix=self.config_prefix(),
                resource_graph=resource_graph,
                output_session=output_session,
                state_session=state_session,
                dependencies=self.config.depends_on,
            )

        return tasks


class Plan(abc.ABC):
    """"""

    task_graph: "ResourceTaskGraph"

    @abc.abstractmethod
    def build_task_graph(self) -> None:
        """
        Call to build and validate the task grpah for this plan. After this is called,
        `self.task_graph` should not be None
        """
        raise NotImplementedError

    @abc.abstractmethod
    def is_empty(self, include_metatasks: bool = False) -> bool:
        """
        Indicate whether this plan is empty, meaning any tasks it contains
        are metatasks or there are no tasks at all
        """
        raise NotImplementedError

    @abc.abstractmethod
    async def plan(
        self,
        config_session: ResourceSession,
        migrator: Optional["Migrator"] = None,
        **kwargs,
    ) -> "Plan":
        """
        Create a plan from the output of this plan and a new session
        """
        raise NotImplementedError


@dc.dataclass(frozen=True)
class DefaultPlan(Plan):
    """
    A plan contains enough information to inspect and execute the sequence
    of tasks to reach the state of config_session from state_graph
    """

    nodes: Sequence[PlanNode]
    providers: Sequence[Provider]
    config_session: ResourceSession
    state_session: ResourceSession
    state_graph: ResourceGraph
    migrator: Optional["Migrator"] = None
    task_graph: "ResourceTaskGraph" = dc.field(init=False, default=None)

    def new_task_graph(self, strict: bool = False) -> "ResourceTaskGraph":
        """
        Render a full task graph from this plan
        """
        from statey.executor import ResourceTaskGraph

        output_session = self.config_session.clone()

        node_dict = {node.key: node for node in self.nodes}

        graphs = []
        raw_edges = {}

        for node in self.nodes:
            graphs.append(
                node.get_task_graph(
                    resource_graph=self.state_graph,
                    output_session=output_session,
                    state_session=self.state_session,
                )
            )
            node_edges = node.get_edges(node_dict)
            for edge_from, edge_to, data in node_edges:
                raw_edges.setdefault((edge_from, edge_to), []).append(data)

        edges = [
            (*key, {"optional": all(item.get("optional") for item in vals)})
            for key, vals in raw_edges.items()
        ]

        # This will raise an error if there are overlapping keys, which there
        # shouldn't be.
        full_graph = reduce(nx.union, graphs) if graphs else nx.DiGraph()
        full_graph.add_edges_from(edges)

        original_graph = full_graph.copy()

        check_dag(full_graph, strict)

        return ResourceTaskGraph(
            full_graph, output_session, self.state_graph, original_graph
        )

    def build_task_graph(self, force: bool = False, **kwargs) -> None:
        """
        Create the task graph for this plan if it does not already exist
        """
        if self.task_graph is not None and not force:
            return

        self.__dict__["task_graph"] = self.new_task_graph(**kwargs)

    def is_empty(self, include_metatasks: bool = False) -> bool:
        """
        Indicate whether this plan is empty, meaning any tasks it contains
        are metatasks or there are no tasks at all
        """
        self.build_task_graph()
        for node in self.task_graph.task_graph.nodes:
            task = self.task_graph.get_task(node)
            if include_metatasks or not task.is_metatask():
                return False
        return True

    async def plan(
        self,
        config_session: ResourceSession,
        migrator: Optional["Migrator"] = None,
        **kwargs,
    ) -> "Plan":
        """
        Create a new plan from the output of this plan to the new `config_session`.
        """
        if migrator is None:
            migrator = self.migrator

        if migrator is None:
            migrator = DefaultMigrator()

        self.build_task_graph()
        new_plan = await migrator.plan(
            config_session,
            self.task_graph.resource_graph,
            self.task_graph.output_session,
            **kwargs,
        )
        compound = CompoundPlan(self, new_plan)
        compound.build_task_graph()
        return compound


@dc.dataclass(frozen=True)
class CompoundPlan(Plan):
    """"""

    first: Plan
    second: Plan
    task_graph: "ResourceTaskGraph" = dc.field(init=False, default=None)

    def _prefix_graph(self, key: Callable[[str], str], graph: nx.DiGraph) -> nx.DiGraph:

        new_graph = nx.DiGraph()

        for node in graph.nodes:
            new_graph.add_node(key(node), **graph.nodes[node])

        translated_edges = [
            (key(left), key(right), graph.edges[left, right])
            for left, right in graph.edges
        ]

        new_graph.add_edges_from(translated_edges)
        return new_graph

    def new_task_graph(self, strict: bool = False) -> "ResourceTaskGraph":
        """
        Render a full task graph from this plan
        """
        from statey.executor import ResourceTaskGraph

        self.first.build_task_graph()
        self.second.build_task_graph()

        graph = nx.DiGraph()

        first_graph = self.first.task_graph.task_graph
        second_graph = self.second.task_graph.task_graph

        first_key = "first:{}".format
        second_key = "second:{}".format

        first_prefixed_graph = self._prefix_graph(first_key, first_graph)
        second_prefixed_graph = self._prefix_graph(second_key, second_graph)

        graph = nx.union(first_prefixed_graph, second_prefixed_graph)
        graph.add_edges_from(first_prefixed_graph.edges)
        graph.add_edges_from(second_prefixed_graph.edges)

        first_end_tasks = {
            node
            for node in first_prefixed_graph.nodes
            if not first_prefixed_graph.succ[node]
        }
        second_start_tasks = {
            node
            for node in second_prefixed_graph.nodes
            if not second_prefixed_graph.pred[node]
        }

        edges = []

        for first_task, second_task in product(first_end_tasks, second_start_tasks):
            edges.append((first_task, second_task, {"optional": False}))

        graph.add_edges_from(edges)

        original_graph = graph.copy()

        check_dag(graph, strict)

        return ResourceTaskGraph(
            task_graph=graph,
            output_session=self.second.task_graph.output_session,
            resource_graph=self.second.task_graph.resource_graph,
            original_task_graph=original_graph,
        )

    def is_empty(self, include_metatasks: bool = False) -> bool:
        return self.first.is_empty() and self.second.is_empty()

    # Same build_task_graph() and plan() behavior as DefaultPlan
    build_task_graph = DefaultPlan.build_task_graph

    plan = DefaultPlan.plan


@dc.dataclass(frozen=True)
class StatefulPlan(Plan):
    """"""

    original_plan: Plan
    output_state: str
    migrator: Optional["Migrator"] = None
    task_graph: "ResourceTaskGraph" = dc.field(init=False, default=None)

    def new_task_graph(self, strict: bool = False) -> "ResourceTaskGraph":
        """
        Render a full task graph from this plan
        """
        from statey.executor import ResourceTaskGraph

        task_graph = self.original_plan.new_task_graph()

        nx_graph = task_graph.task_graph

        heads = {node for node in nx_graph.nodes if not nx_graph.pred[node]}

        before_task = SnapshotSwitch(
            task_graph.resource_graph, task_graph.output_session
        )
        before_task_name = "_snapshot:input"

        state_task = SetGraphState(task_graph.resource_graph, self.output_state)
        state_task_name = "state"

        nx_graph.add_node(before_task_name, task=before_task)
        nx_graph.add_node(state_task_name, task=state_task)
        nx_graph.add_edge(before_task_name, state_task_name)
        for node in heads:
            nx_graph.add_edge(state_task_name, node)

        tails = {node for node in nx_graph.nodes if not nx_graph.succ[node]}

        after_task = StateSnapshot(task_graph.resource_graph)
        after_task_name = "_snapshot:state"

        nx_graph.add_node(after_task_name, task=after_task)
        for node in tails:
            nx_graph.add_edge(node, after_task_name)

        return ResourceTaskGraph(
            nx_graph, task_graph.output_session, task_graph.resource_graph
        )

    def is_empty(self, include_metatasks: bool = False) -> bool:
        if include_metatasks:
            return False
        return self.original_plan.is_empty()

    # Same build_task_graph() and plan() behavior as DefaultPlan
    build_task_graph = DefaultPlan.build_task_graph

    plan = DefaultPlan.plan



from statey.plan.actions import PlanAction
from statey.plan.migrators import Migrator, DefaultMigrator, StatefulMigrator
from statey.plan.plans import Plan, DefaultPlan, PlanNode, StatefulPlan, CompoundPlan



import abc
import dataclasses as dc
from typing import Sequence

import networkx as nx

import statey as st
from statey.resource import ResourceGraph, ResourceState, Resource
from statey.syms import session, Object
from statey.task import (
    TaskSession,
    SessionSwitch,
    SessionSwitchAction,
    GraphSetKey,
    GraphDeleteKey,
)


class PlanAction(abc.ABC):
    """
    A plan action is some subset of a full task graph for a given name.
    """

    @abc.abstractmethod
    def render(
        self,
        graph: nx.DiGraph,
        prefix: str,
        resource_graph: ResourceGraph,
        output_session: session.Session,
        state_session: session.Session,
        # Dependencies of the configuration
        dependencies: Sequence[str],
    ) -> None:
        """
        Add the tasks in this graph to the given graph. Note that the output
        node must be {prefix}:output
        """
        raise NotImplementedError

    @abc.abstractmethod
    def input_task(self, prefix: str) -> str:
        """
        Return the name of the task, given a prefix, that will be the "input" task when
        rendering this action
        """
        raise NotImplementedError

    @abc.abstractmethod
    def output_task(self, prefix: str) -> str:
        """
        Return the name of the task, given a prefix, that will be the "output" task when
        rendering this action
        """
        raise NotImplementedError


@dc.dataclass(frozen=True)
class ExecuteTaskSession(PlanAction):
    """"""

    task_session: TaskSession
    task_input_key: str
    task_state_key: str
    output_ref: Object
    graph_ref: Object
    output_key: str
    config_state: ResourceState
    previous_state: ResourceState
    resource: Resource
    input_symbol: Object
    state_symbol: Object
    state_allow_unknowns: bool = False

    def input_task(self, prefix: str) -> str:
        return f"{prefix}:input"

    def output_task(self, prefix: str) -> str:
        return f"{prefix}:output"

    def render(
        self,
        graph: nx.DiGraph,
        prefix: str,
        resource_graph: ResourceGraph,
        output_session: session.Session,
        state_session: session.Session,
        dependencies: Sequence[str],
    ) -> None:

        input_switch_task = SessionSwitch(
            [
                SessionSwitchAction(
                    input_session=output_session,
                    input_symbol=self.input_symbol,
                    output_session=self.task_session,
                    output_key=self.task_input_key,
                ),
                SessionSwitchAction(
                    input_session=state_session,
                    input_symbol=self.state_symbol,
                    output_session=self.task_session,
                    output_key=self.task_state_key,
                    allow_unknowns=self.state_allow_unknowns,
                ),
            ]
        )

        input_key = self.input_task(prefix)
        graph.add_node(input_key, task=input_switch_task, source=prefix)

        # If output state is null, our graph operation is a deletion. Otherwise, it's a "set"
        if self.config_state == self.resource.null_state:
            output_key = self.output_task(prefix)
            output_graph_task = GraphDeleteKey(
                key=self.output_key,
                resource_graph=resource_graph,
                input_session=self.task_session,
                input_symbol=self.graph_ref,
            )
            graph.add_node(output_key, task=output_graph_task, source=prefix)
        else:
            output_key = f"{prefix}:state"
            output_graph_task = GraphSetKey(
                input_session=self.task_session,
                input_symbol=self.graph_ref,
                dependencies=dependencies,
                key=self.output_key,
                resource_graph=resource_graph,
                state=self.config_state,
                finalize=self.resource.finalize,
            )
            graph.add_node(output_key, task=output_graph_task, source=prefix)

            output_switch_task = SessionSwitch(
                [
                    SessionSwitchAction(
                        input_session=self.task_session,
                        input_symbol=self.output_ref,
                        output_session=output_session,
                        output_key=self.output_key,
                    )
                ]
            )
            output_switch_key = self.output_task(prefix)
            graph.add_node(output_switch_key, task=output_switch_task, source=prefix)

            # We should always update the state before the session
            graph.add_edge(output_key, output_switch_key)

        task_graph = self.task_session.task_graph(resource_graph, self.output_key)
        if not task_graph.nodes:
            graph.add_edge(input_key, output_key)

        task_key = lambda x: f"{prefix}:task:{x}"

        for sub_node in nx.topological_sort(task_graph):
            node_key = task_key(sub_node)
            task = task_graph.nodes[sub_node]["task"]

            graph.add_node(node_key, task=task, source=prefix)

            # If not predecessors, make it dependent on the input
            # switch
            if not task_graph.pred[sub_node]:
                graph.add_edge(input_key, node_key)
            # Otherwise, add upstream edges
            else:
                for pred in task_graph.pred[sub_node]:
                    graph.add_edge(task_key(pred), node_key)

            # The opposite of above--if no dependencies, make the
            # output a dependency
            if not task_graph.succ[sub_node]:
                graph.add_edge(node_key, output_key)


@dc.dataclass(frozen=True)
class SetValue(PlanAction):
    """
    Set a given value in either the resource graph, the output session, or both during execution
    """

    output_key: str
    output_symbol: Object
    set_graph: bool = True
    set_session: bool = True

    def __post_init__(self) -> None:
        if not (self.set_graph or self.set_session):
            raise ValueError(
                "At least one of `set_session` or `set_graph` must be true."
            )

    def graph_update_task(self, prefix: str) -> str:
        return f"{prefix}:state"

    def session_switch_task(self, prefix: str) -> str:
        return f"{prefix}:output"

    def input_task(self, prefix: str) -> str:
        if self.set_graph and self.set_session or not self.set_session:
            return self.graph_update_task(prefix)
        return self.session_switch_task(prefix)

    def output_task(self, prefix: str) -> str:
        if self.set_session:
            return self.session_switch_task(prefix)
        return self.graph_update_task(prefix)

    def render(
        self,
        graph: nx.DiGraph,
        prefix: str,
        resource_graph: ResourceGraph,
        output_session: session.Session,
        state_session: session.Session,
        dependencies: Sequence[str],
    ) -> None:

        graph_task_name = self.graph_update_task(prefix)
        if self.set_graph:
            graph_task = GraphSetKey(
                input_session=output_session,
                key=self.output_key,
                input_symbol=self.output_symbol,
                resource_graph=resource_graph,
                dependencies=dependencies,
            )
            graph.add_node(graph_task_name, task=graph_task, source=prefix)

        session_task_name = self.session_switch_task(prefix)
        if self.set_session:
            session_task = SessionSwitch(
                [
                    SessionSwitchAction(
                        input_session=output_session,
                        input_symbol=self.output_symbol,
                        output_session=output_session,
                        output_key=self.output_key,
                    )
                ]
            )
            graph.add_node(session_task_name, task=session_task, source=prefix)
            if self.set_graph:
                graph.add_edge(graph_task_name, session_task_name)


@dc.dataclass(frozen=True)
class ResourceSetValue(PlanAction):
    """
    SetValue for a resource, a no-op action
    """

    output_key: str
    output_symbol: Object
    resource: Resource
    state: ResourceState

    def graph_update_task(self, prefix: str) -> str:
        return f"{prefix}:state"

    def session_switch_task(self, prefix: str) -> str:
        return f"{prefix}:output"

    def input_task(self, prefix: str) -> str:
        return None
        # return self.graph_update_task(prefix)

    def output_task(self, prefix: str) -> str:
        return self.session_switch_task(prefix)

    def render(
        self,
        graph: nx.DiGraph,
        prefix: str,
        resource_graph: ResourceGraph,
        output_session: session.Session,
        state_session: session.Session,
        dependencies: Sequence[str],
    ) -> None:

        session_task_name = self.session_switch_task(prefix)
        session_task = SessionSwitch(
            [
                SessionSwitchAction(
                    input_session=output_session,
                    input_symbol=self.output_symbol,
                    output_session=output_session,
                    output_key=self.output_key,
                )
            ]
        )
        graph.add_node(session_task_name, task=session_task, source=prefix)

        graph_task_name = self.graph_update_task(prefix)
        graph_task = GraphSetKey(
            input_session=output_session,
            key=self.output_key,
            input_symbol=self.output_symbol,
            resource_graph=resource_graph,
            dependencies=dependencies,
            state=self.state,
            finalize=self.resource.finalize,
        )
        graph.add_node(graph_task_name, task=graph_task, source=prefix)

        graph.add_edge(graph_task_name, session_task_name)


@dc.dataclass(frozen=True)
class DeleteValue(PlanAction):
    """"""

    delete_key: str

    def input_task(self, prefix: str) -> str:
        return f"{prefix}:state"

    def output_task(self, prefix: str) -> str:
        return f"{prefix}:state"

    def render(
        self,
        graph: nx.DiGraph,
        prefix: str,
        resource_graph: ResourceGraph,
        output_session: session.Session,
        state_session: session.Session,
        dependencies: Sequence[str],
    ) -> None:

        graph_task_name = self.input_task(prefix)
        graph_task = GraphDeleteKey(key=self.delete_key, resource_graph=resource_graph)
        graph.add_node(graph_task_name, task=graph_task, source=prefix)



# import abc
# import dataclasses as dc
# from itertools import product
# from typing import Optional, Callable, Any, Dict, Sequence

# import networkx as nx

# import statey as st
# from statey.plan import Plan, StatefulMigrator
# from statey.resource import ResourceGraph, DefaultResourceGraph
# from statey.syms import session, types, utils


# class StatefulResourceGraph(ResourceGraph):
#     """
#     Resource graph that exposes two additional methods--get_ and set_ methods for the state
#     """

#     @abc.abstractmethod
#     def get_state(self) -> str:
#         """
#         Get the state name for this resource graph
#         """
#         raise NotImplementedError

#     @abc.abstractmethod
#     def set_state(self, state: str) -> None:
#         """
#         Set the state name for this resource graph
#         """
#         raise NotImplementedError


# @dc.dataclass(frozen=True)
# class DefaultStatefulResourceGraph(StatefulResourceGraph):
#     """
#     Default implementation for a stateful resource graph
#     """

#     graph: ResourceGraph
#     state: dc.InitVar[str] = "DOWN"
#     _state: str = dc.field(init=False, default=None)

#     def __post_init__(self, state: str) -> None:
#         self.set_state(state)

#     def get_state(self) -> str:
#         return self._state

#     def set_state(self, state: str) -> None:
#         self.__dict__["_state"] = state

#     def get(self, key: str) -> Dict[str, Any]:
#         return self.graph.get(key)

#     def keys(self) -> Sequence[str]:
#         return self.graph.keys()

#     def add_dependencies(self, key: str, dependencies: Sequence[str]) -> None:
#         self.graph.add_dependencies(key, dependencies)

#     async def refresh(
#         self, registry: Optional[st.Registry] = None, finalize: bool = False
#     ) -> Iterator[str]:
#         async for item in self.graph.refresh(registry, finalize):
#             yield item

#     def clone(self) -> "ResourceGraph":
#         return type(self)(graph=self.graph.clone(), state=self.get_state())

#     def to_dict(self, registry: Optional["Registry"] = None) -> Dict[str, Any]:
#         return {"state": self.get_state(), "resources": self.graph.to_dict(registry)}

#     @classmethod
#     def from_dict(
#         cls, data: Any, registry: Optional["Registry"] = None
#     ) -> "ResourceGraph":
#         graph = DefaultResourceGraph.from_dict(data["resources"], registry)
#         return cls(graph=graph, state=data["state"])

#     def to_session(self) -> ResourceSession:
#         return self.graph.to_session()


# class ModuleTransition:
#     """"""

#     func: Callable[[session.Session], None]
#     to_state: str
#     from_state: Optional[str] = None

#     def _get_session(self, states: Dict[str, types.Type]) -> session.Session:
#         """"""
#         session = st.create_resource_session()
#         if self.from_state is not None:
#             snapshot_type = states[self.from_state]
#             session["_snapshot"] << st.Unknown[snapshot_type]

#         self.func(session)
#         return session.ns.root._type

#     def session(self, states: Dict[str, types.Type]) -> session.Session:
#         """"""
#         return self._get_session(states)


# def build_state_graph(
#     transitions: Sequence[ModuleTransition], registry: Optional["Registry"] = None
# ) -> nx.DiGraph:
#     """"""
#     if registry is None:
#         registry = st.registry

#     graph = nx.DiGraph()
#     edges = set()
#     from_any_state = set()

#     # First just build the graph nodes, then add edges
#     for transition in transitions:
#         if transition.to_state in graph.nodes:
#             graph.nodes[transition.to_state]["transitions"].append(transition)
#         else:
#             graph.add_node(transition.to_state, transitions=[transition])
#         if transition.from_state is not None:
#             edges.add((transition.from_state, transition.to_state))
#         else:
#             from_any_state.add(transition.to_state)

#     graph.add_edges_from(edges)

#     states = {}

#     for node in nx.topological_sort(graph):
#         node_transitions = graph.nodes[node]["transitions"]

#         stable_output_type = None
#         stateful_transitions = []

#         for transition in node_transitions:
#             if transition.from_state is None:
#                 stable_output_type = transition.session(states).ns.root._type
#             else:
#                 stateful_transitions.append(transition)

#         if stable_output_type is None:
#             raise ValueError(f"{node} does not have a stable transition!")

#         invalid_transitions = []
#         for transition in stateful_transitions:
#             output_type = transition.session(states).ns.root._type
#             try:
#                 registry.get_caster(output_type, stable_output_type)
#             except st.exc.NoCasterFound:
#                 invalid_transitions.append(transition)

#         if invalid_transitions:
#             raise ValueError(
#                 f"{node}: transitions have invalid output types: {invalid_transitions}."
#             )

#         states[node] = graph.nodes[node]["type"] = stable_output_type

#     for from_node, to_node in product(graph.nodes, from_any_state):
#         if from_node == to_node:
#             continue
#         graph.add_edge(from_node, to_node)

#     return graph


# class Module(abc.ABC):
#     """"""

#     def __init__(self, transitions: Sequence[ModuleTransition]) -> None:
#         self.transitions = transitions
#         self.graph = build_state_graph(transitions)

#     def migrator(self) -> StatefulMigrator:
#         return StatefulMigrator()

#     async def plan(self, resource_graph: StatefulResourceGraph, desired: str) -> Plan:

#         current_state = resource_graph.get_state()

#         path = [current_state] + nx.shortest_path(current_state, desired, self.graph)

#         # First collect a list of transitions detailing the path
#         transitions = []
#         transition_steps = []

#         for from_state, to_state in zip(path, path[1:]):
#             node_transitions = self.graph.node[to_state]["transitions"]

#             added = False
#             stable = None

#             for transition in node_transitions:
#                 if transition.from_state == from_state:
#                     transitions.append(transition)
#                     added = True
#                     break
#                 elif transition.from_state is None:
#                     stable = transition

#             if not added:
#                 transitions.append(stable)

#             transition_steps.append(to_state)

#         head, *tail = transitions
#         head_state, *tail_states = transition_steps

#         migrator = self.migrator()

#         states = {node: self.graph.nodes[node]["type"] for node in self.graph.nodes}

#         head_session = head.session(states)

#         plan = await migrator.plan(head_session, resource_graph, state=head_state)

#         for transition, tail_state in zip(tail, tail_states):
#             session = transition.session(states)
#             plan = await plan.plan(session, state=tail_state)

#         return plan



try:
    import graphviz
except ImportError as err:
    raise RuntimeError(
        f"`graphviz` is not installed, this extension module cannot be used."
    ) from err

import dataclasses as dc
from typing import Dict, Any, Callable, Optional

from graphviz import Digraph

import statey as st
from statey.syms import utils


class TaskGraphRenderer:
    """
    Defines some methods for customizing how a ResourceTaskGraph is converted
    to a graphviz.Digraph
    """

    def node_key(self, key: str) -> str:
        return key.replace(":", "/")

    def node_args(self, key: str) -> Dict[str, Any]:
        return {}

    def edge_args(
        self, from_key: str, to_key: str, exists: bool, data: Dict[str, Any]
    ) -> Dict[str, Any]:
        color = "green"
        kws = {}
        if not exists:
            color = "red"
            kws["constraint"] = "false"
        elif data.get("optional"):
            color = "gold2"
        return {"color": color, **kws}


def get_graphviz_graph(
    task_graph: st.ResourceTaskGraph,
    renderer: TaskGraphRenderer = TaskGraphRenderer(),
    task_filter: Optional[Callable[[str, st.Task], bool]] = None,
    **kwargs: Dict[str, Any],
) -> Digraph:
    """
    Get a graphviz Digraph from the given task graph that can be rendered into
    an image
    """
    dot = Digraph(**kwargs)

    original_task_graph = task_graph.original_task_graph
    real_task_graph = task_graph.task_graph
    if original_task_graph is None:
        original_task_graph = real_task_graph

    if task_filter is not None:
        keep = set()

        for node in real_task_graph.nodes:
            task = task_graph.get_task(node)
            if task_filter(node, task):
                keep.add(node)

        utils.subgraph_retaining_dependencies(original_task_graph, keep)
        utils.subgraph_retaining_dependencies(real_task_graph, keep)

    for node in original_task_graph.nodes:
        args = renderer.node_args(node)
        dot.node(renderer.node_key(node), **args)

    for node_from, node_to in original_task_graph.edges:
        exists = (node_from, node_to) in real_task_graph.edges
        data = original_task_graph.edges[node_from, node_to]
        args = renderer.edge_args(node_from, node_to, exists, data)
        dot.edge(renderer.node_key(node_from), renderer.node_key(node_to), **args)

    return dot
